# -*- coding: utf-8 -*-
"""
Provides the base class for specifying the source of a lensing system.

This module contains the base class that all the source classes will build
from. Because the steps for rendering a source can vary between different
models, the required functions are very sparse.
"""
import copy
from ..Utils.cosmology_utils import get_cosmology


class SourceBase():
	"""
	Base class for producing lenstronomy LightModel arguments

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): dictionary with source-specific parameters
	"""

	required_parameters = tuple()

	def __init__(self, cosmology_parameters, source_parameters):
		self.cosmo = get_cosmology(cosmology_parameters)
		self.source_parameters = copy.deepcopy(source_parameters)

		# Check that all the required parameters are present
		self.check_parameterization(self.__class__.required_parameters)

	def check_parameterization(self, required_params):
		""" Check that all the required parameters are present in the
		source_parameters.

		Args:
			required_params ([str,...]): A list of strings containing the
				required parameters.
		"""
		if not all(elem in self.source_parameters.keys() for
			elem in required_params):
			raise ValueError('Not all of the required parameters for the ' +
				'parameterization are present.')

	def update_parameters(self, cosmology_parameters=None,source_parameters=None):
		"""Updated the class parameters

		Args:
			cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
				Either a name of colossus cosmology, a dict with 'cosmology name':
				name of colossus cosmology, an instance of colussus cosmology, or
				a dict with H0 and Om0 ( other parameters will be set to
				defaults).
			source_parameters (dict): A dictionary containing all the parameters
				needed to draw sources.
		"""
		if source_parameters is not None:
			self.source_parameters.update(source_parameters)
		if cosmology_parameters is not None:
			self.cosmo = get_cosmology(cosmology_parameters)

	def draw_source(self):
		"""Return lenstronomy LightModel names and kwargs

		Returns:
			(list,list,list) A list containing the model name(s),
			a list containing the model kwargs dictionaries, and a list
			containing the redshifts of each model. Redshifts list can
			be None.
		"""
		raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Provides the base class for specifying the point source of a lensing system.

This module contains the base class that all the point source classes will 
build from. Because the steps for rendering a source can vary between different
models, the required functions are very sparse.
"""
import copy


class PointSourceBase:
	"""
	Base class for producing lenstronomy PointSource arguments

	Args:
		point_source_parameters (dict): dictionary with source-specific 
			parameters.

	Notes:
		Has no required parameters by default.
	"""

	required_parameters = tuple()

	def __init__(self, point_source_parameters):
		self.point_source_parameters = copy.deepcopy(point_source_parameters)

		# Check that all the required parameters are present
		self.check_parameterization(self.__class__.required_parameters)

	def check_parameterization(self, required_params):
		""" Check that all the required parameters are present in the
		point_source_parameters.

		Args:
			required_params ([str,...]): A list of strings containing the
				required parameters.
		"""
		if not all(elem in self.point_source_parameters.keys() for
			elem in required_params):
			raise ValueError('Not all of the required parameters for the ' +
				'parameterization are present.')

	def update_parameters(self, point_source_parameters=None):
		"""Update the class parameters

		Args:
			point_source_parameters (dict): A dictionary containing all the 
				parameters needed to draw point sources.
		"""
		if point_source_parameters is not None:
			self.point_source_parameters.update(point_source_parameters)

	def draw_point_source(self):
		"""Return lenstronomy PointSource names and kwargs

		Returns:
			(list,list) A list containing the model name(s), and
			a list containing the model kwargs dictionaries.
		"""
		raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Add to the functionality of colossus

Useful functions for extra cosmology calculations.
"""
from colossus.cosmology import cosmology
import numpy as np


def get_cosmology(cosmology_parameters):
	"""Returns colossus cosmology

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or
			a dict with H0 and Om0 ( other parameters will be set to
			defaults).

	Returns:
		(colossus.cosmology.cosmology.Cosmology): A corresponding instance
		of the colossus cosmology class.
	"""
	if isinstance(cosmology_parameters, cosmology.Cosmology):
		return cosmology_parameters
	if isinstance(cosmology_parameters, str):
		return cosmology.setCosmology(cosmology_parameters)
	if isinstance(cosmology_parameters, dict):
		if 'cosmology_name' in cosmology_parameters:
			return get_cosmology(cosmology_parameters['cosmology_name'])
		else:
			# Leave some parameters to their default values so the user only
			# has to specify H0 and Om0.
			col_params = dict(flat=True, H0=cosmology_parameters['H0'],
				Om0=cosmology_parameters['Om0'],
				Ob0=cosmology_parameters['Ob0'],
				sigma8=cosmology_parameters['sigma8'],
				ns=cosmology_parameters['ns'])
			return cosmology.setCosmology('temp_cosmo', col_params)


def kpc_per_arcsecond(z,cosmo):
	"""Calculates the physical kpc per arcsecond at a given redshift and
	cosmology

	Args:
		z (float): The redshift to calculate the distance at
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		(float): The kpc per arcsecond
	"""
	h = cosmo.h
	kpc_per_arcsecond = (cosmo.angularDiameterDistance(z) *np.pi/180/3600 /
		h * 1e3)
	return kpc_per_arcsecond


def ddt(sample,cosmo):
	"""Calculates time delay distance given lens redshift, source redshift, and
	cosmology.

	Args: 
		sample (dict): Dictionary containing dictionaries of parameters for each
			model component. Generated using Sampler .sample() method
		cosmo (colossus.cosmology.Cosmology): An instance of the colossus
			cosmology object
	
	Returns:
		(float): Time delay distance
	"""
	z_lens = sample['main_deflector_parameters']['z_lens']
	z_source = sample['source_parameters']['z_source']
	D_d = cosmo.angularDiameterDistance(z_lens)
	D_s = cosmo.angularDiameterDistance(z_source)
	D_ds =  (1/ (1+z_source)) * cosmo.comovingDistance(z_min=z_lens,
		z_max=z_source)
	# convert from Mpc/h to Mpc 
	return (1+z_lens) * D_d * D_s / (D_ds*cosmo.h)



# -*- coding: utf-8 -*-
"""
Define the sampler class which reads goes from the distribution dictionary to
drawing samples on the lens parameters.

This module contains the class used to sample parameters for our train and test
set from the input distributions.
"""
import warnings
# Definte the components we need the sampler to consider.
# TODO: add point source & lens light
lensing_components = ['subhalo','los','main_deflector','source','lens_light',
	'point_source','lens_equation_solver','cosmology','psf','detector','drizzle']

# Global filters on the python warnings. Using this since filter
# behaviour is a bit weird.
CROSSOBJECTWARNING = True


class Sampler():
	"""Class for drawing lens parameter values from input distribution
	dictionaries

	Args:
		configuration_dictionary (dict): An instance of the configuration
			dictionary that will be used to decide how to sample parameter
			values.
	Notes:
		For examples on how to configure the dict object, see the
		ConfigurationDict.ipynb example notebook.
	"""

	def __init__(self,configuration_dictionary):
		self.config_dict = configuration_dictionary

	@staticmethod
	def draw_from_dict(draw_dict):
		"""Populates a dict with samples drawn from the specified distributions
		in the input dict.

		Args:
			draw_dict (dict): The dictionary containing keys mapping to values
				or distributions for each parameter.

		Returns:
			(dict): A dict with a drawn value for each parameter.

		Notes:
			Multivariate distribution for parameters should have a key of the
			form 'param_1,param_2,param_3'.
		"""
		param_dict = {}

		# Iterate through the keys in the draw_dict and populate the values of
		# param_dict correctly.
		for key in sorted(draw_dict):
			# If the key implies that multiple parameters will be drawn from
			# the distribution, draw the value and then iterate through the
			# parameters.
			if ',' in key:
				# Get the parameters, removing whitespace
				params = key.replace(' ','').split(',')
				# Draw the values
				draw = draw_dict[key]()
				# Check for consistency
				if len(params) != len(draw):
					raise ValueError('Parameters of length %d do'%(len(params))
						+ ' not match draw of length %d'%(len(draw)))
				# Populate the keys
				for i, param in enumerate(params):
					param_dict[param] = draw[i]
			# If it's a univariate function just call it.
			elif callable(draw_dict[key]):
				param_dict[key] = draw_dict[key]()
			# If it's a fixed value just populate it.
			else:
				param_dict[key] = draw_dict[key]

		return param_dict

	def sample(self):
		"""Samples from the distributions given in the configuration
		dictionary

		Returns:
			(dict): A dictionary containing the parameter values that will
			be sampled.
		"""
		# Pull the global warning variable and initialize our dict
		global CROSSOBJECTWARNING
		full_param_dict = {}

		# For each possible component of our lensing add the parameters
		for component in lensing_components:
			if component in self.config_dict:
				draw_dict = self.config_dict[component]['parameters']
				param_dict = self.draw_from_dict(draw_dict)
				full_param_dict[component+'_parameters'] = param_dict

		# Populate parameters from distributions that span accross objects.
		if 'cross_object' in self.config_dict:
			draw_dict = self.config_dict['cross_object']['parameters']
			cross_dict = self.draw_from_dict(draw_dict)
			# Go through the params and update the full dict
			for cross_param in cross_dict:
				component, param = cross_param.split(':')
				param_dict = full_param_dict[component+'_parameters']
				# Warn the user the first time an overwrite happens
				if param in param_dict and CROSSOBJECTWARNING:
					warnings.warn('Parameter in cross dict specified elsewhere!'
						+ ' Will be overwritten')
					CROSSOBJECTWARNING = False
				full_param_dict[component+'_parameters'][param] = (
					cross_dict[cross_param])

		# Populate the cross objects
		return full_param_dict



# -*- coding: utf-8 -*-
"""
Define a number of distributions that can be used by the sampler to draw
lensing parameters.

This module contains classes that define distributions that can be effecitvely
sampled from.
"""
import numpy as np
from scipy.stats import truncnorm


class MultivariateLogNormal():
	"""Class for drawing multiple parameters from a multivariate log normal
	distribution.

	Args:
		mean (np.array): The mean value of the log normal distribution
		covariance (np.array): The covariance matrix of the distribution

	Notes:
		Returned values will follow the distribution exp(normal(mu,covariance))
	"""

	def __init__(self,mean,covariance):
		# Make sure the dimension of the mean allows for easy matrix
		# operations
		if len(mean.shape)==1:
			mean = np.expand_dims(mean,axis=0)
		self.mean = mean
		self.covariance = covariance
		self.L = np.linalg.cholesky(self.covariance)

	def __call__(self,n=1):
		"""Returns n draws from the distribution

		Args:
			n (int): The number of draws.

		Returns
			(np.array): n x len(mean) array of draws.
		"""
		rand_draw = np.random.randn(self.mean.shape[1]*n).reshape(
			(self.mean.shape[1],n))
		return np.exp(self.mean.T+np.dot(self.L,rand_draw)).T


class TruncatedMultivariateNormal():
	"""Class for drawing multiple parameters from a truncated multivariate
	normal distribution.

	Args:
		mean (np.array): The mean value of the log normal distribution
		covariance (np.array): The covariance matrix of the distribution
		min_values (np.array): The minimum value for each parameter
		max_values (np.array): The maximum value for each parameter

	Notes:
		Note this code uses rejection sampling, which may be slow if most
		of the multivariate normal's pdf is outside the bounds. Be careful when
		using this in high dimensions.
	"""

	def __init__(self,mean,covariance,min_values=None,max_values=None):
		# Make sure that each of the n-dimensional inputs follows
		# the desired shape for matrix calculations
		if len(mean.shape)==1:
			mean = np.expand_dims(mean,axis=0)
		# If none for min_values, set to negative infinity
		if min_values is None:
			min_values = np.ones(mean.shape) * -np.inf
		elif len(min_values.shape)==1:
			min_values = np.expand_dims(min_values,axis=0)
		# If none for max_values, set to positive infinity
		if max_values is None:
			max_values = np.ones(mean.shape) * np.inf
		elif len(max_values.shape)==1:
			max_values = np.expand_dims(max_values,axis=0)

		self.mean = mean
		self.covariance = covariance
		self.L = np.linalg.cholesky(self.covariance)
		self.min_values = min_values
		self.max_values = max_values

	def __call__(self,n=1):
		"""Returns n draws from the distribution

		Args:
			n (int): The number of draws.

		Returns
			(np.array): n x len(mean) array of draws.
		"""
		# Start with a regular draw
		n_accepted = 0
		n_samp = n
		keep_draws = np.zeros((n,self.mean.shape[1]))
		rand_draw = np.random.randn(self.mean.shape[1]*n_samp).reshape(
			(self.mean.shape[1],n_samp))
		draws = (self.mean.T+np.dot(self.L,rand_draw)).T

		# Check which draws are within our bounds
		keep_ind = np.prod(draws > self.min_values,axis=-1,dtype=np.bool)
		keep_ind *= np.prod(draws < self.max_values,axis=-1,dtype=np.bool)
		keep_draws[n_accepted:n_accepted+np.sum(keep_ind)] = draws[keep_ind]
		n_accepted += np.sum(keep_ind)

		# Keep drawing until we have enough samples.
		while n_accepted<n:
			# Draw
			rand_draw = np.random.randn(self.mean.shape[1]*n_samp).reshape(
				(self.mean.shape[1],n_samp))
			draws = (self.mean.T+np.dot(self.L,rand_draw)).T
			# Check for the values in bounds
			keep_ind = np.prod(draws > self.min_values,axis=-1,dtype=np.bool)
			keep_ind *= np.prod(draws < self.max_values,axis=-1,dtype=np.bool)
			# Only use the ones we need
			use_keep = np.minimum(n-n_accepted,np.sum(keep_ind))
			keep_draws[n_accepted:n_accepted+use_keep] = (
				draws[keep_ind][:use_keep])
			n_accepted += use_keep

		# Remove first dimension for the n=1 case
		if n==1:
			keep_draws = np.squeeze(keep_draws)

		return keep_draws


class EllipticitiesTranslation():
	"""Class that takes in distributions for q_lens and phi_lens, returns 
	samples of e1 and e2 correspondingly
	
	Args: 
		q_dist (scipy.stats.rv_continuous.rvs or float): distribution for 
			axis ratio (can be callable or constant)
		phi_dist (scipy.stats.rv_continuous.rvs or float): distribution for 
			orientation angle in radians (can be callable or constant)

	Notes: 

	"""
	def __init__(self,q_dist,phi_dist):
		self.q_dist = q_dist
		self.phi_dist = phi_dist

	def __call__(self):
		"""Returns a sample of e1,e2 

		Returns:
			(float,float): samples of x-direction ellipticity 
				eccentricity, xy-direction ellipticity eccentricity
		"""

		if callable(self.q_dist):
			q = self.q_dist()
		else:
			q = self.q_dist
		if callable(self.phi_dist):
			phi = self.phi_dist()
		else:
			phi = self.phi_dist
		
		e1 = (1 - q)/(1+q) * np.cos(2*phi)
		e2 = (1 - q)/(1+q) * np.sin(2*phi)

		return e1,e2


class ExternalShearTranslation():
	"""Class that maps samples of gamma_ext, phi_ext distributions to 
		gamma1, gamma2
	
	Args: 
		gamma_dist (scipy.stats.rv_continuous.rvs or float): distribution for 
			external shear modulus (callable or constant) 
		phi_dist (scipy.stats.rv_continuous.rvs or float): distribution for 
			orientation angle in radians (callable or constant)
	
	Notes:
	"""

	def __init__(self, gamma_dist,phi_dist):
		self.gamma_dist = gamma_dist
		self.phi_dist = phi_dist
	
	def __call__(self):
		"""Returns gamma1, gamma2 samples

		Returns:
			(float,float): samples of external shear coordinate values 
		"""
		if callable(self.gamma_dist):
			gamma = self.gamma_dist()
		else:
			gamma = self.gamma_dist
		if callable(self.phi_dist):
			phi = self.phi_dist()
		else:
			phi = self.phi_dist
		
		gamma1 = gamma * np.cos(2*phi)
		gamma2 = gamma * np.sin(2*phi)
		
		return gamma1,gamma2


class KappaTransformDistribution():
	"""Class that samples Kext given 1 / (1-Kext) ~ n. n is sampled from a 
	distribution given by n_dist, then Kext is computed given the
	transformation
	
	Args: 
		n_dist (scipy.stats.rv_continuous.rvs or float): distribution for 
			1 / (1-Kext) (can be callable or constant)
	"""

	def __init__(self,n_dist):
		self.n_dist = n_dist

	def __call__(self):
		"""Samples 1/(1-Kext), then maps that sample to Kext value

		Returns: 
			(float): Kext sample
		"""
		if callable(self.n_dist):
			n = self.n_dist()
		else:
			n = self.n_dist
		
		return 1 - (1/n)


class Duplicate():
	"""Class that returns two copies of the same random draw.

	Args:
		dist (scipy.stats.rv_continuous.rvs or float): The distribution to
			draw the sample from.
	"""

	def __init__(self,dist):
		self.dist = dist

	def __call__(self):
		"""Returns two copies of the same sample

		Returns
			(float,float): Two copies of the sample.
		"""

		if callable(self.dist):
			samp = self.dist()
		else:
			samp = self.dist

		return samp,samp


class DuplicateXY():
	"""Class that returns two copies of x, y coordinates drawn from 
		distributions

	Args: 
		x_dist (scipy.stats.rv_continuous.rvs or float): distribution for x 
			(can be callable or constant)
		y_dist (scipy.stats.rv_continuous.rvs or float): distribution for y 
			(can be callable or constant)
	"""

	def __init__(self,x_dist,y_dist):
		self.x_dist = x_dist
		self.y_dist = y_dist
	
	def __call__(self):
		"""Returns two copies of x,y sample

		Returns
			(float,float,float,float): Two copies of x,y sampled from x_dist
				and y_dist
		"""

		if callable(self.x_dist):
			x = self.x_dist()
		else:
			x = self.x_dist
		if callable(self.y_dist):
			y = self.y_dist()
		else:
			y = self.y_dist
		
		return x,y,x,y


class RedshiftsTruncNorm():
	"""Class that samples z_lens and z_source from truncated normal 
		distributions, forcing z_source > z_lens to be true

	Args: 
		z_lens_min (float): minimum allowed lens redshift
		z_lens_mean (float): lens redshift mean
		z_lens_std (float): lens redshift standard deviation
		z_source_min (float): minimum allowed source redshift
		z_source_mean (float): source redshift mean
		z_source_std (float): source redshift standard deviation
	"""

	def __init__(self, z_lens_min,z_lens_mean,z_lens_std,z_source_min,
		z_source_mean,z_source_std):
		# transform z_lens_min, z_source_min to be in units of std. deviations
		self.z_lens_min = (z_lens_min - z_lens_mean) / z_lens_std 
		self.z_source_min = (z_source_min - z_source_mean) / z_source_std
		# define truncnorm dist for lens redshift
		self.z_lens_dist = truncnorm(self.z_lens_min,np.inf,loc=z_lens_mean,
			scale=z_lens_std).rvs
		# save z_source info
		self.z_source_mean = z_source_mean
		self.z_source_std = z_source_std
	
	def __call__(self):
		"""Returns samples of redshifts, ensuring z_source > z_lens

		Returns: 
			(float,float): z_lens,z_source 
		"""
		z_lens = self.z_lens_dist()
		clip = (z_lens - self.z_source_mean) / self.z_source_std
		# number of std. devs away to stop (negative)
		if(clip > self.z_source_min):
			self.z_source_min = clip
		# define truncnorm dist for source redshift
		z_source = truncnorm(self.z_source_min,np.inf,self.z_source_mean,
			self.z_source_std).rvs()

		return z_lens,z_source


class MultipleValues():
	"""Class to call dist.rvs(size=num)

	Args:
		dist (scipy.stats.rv_continuous.rvs): callable distribution
		num (int): number of samples to return in one call
	"""

	def __init__(self, dist, num):
		self.dist = dist
		self.num = num
	
	def __call__(self):
		"""Returns specified # of samples from dist
		
		Returns: 
			list(float): |num| samples from dist
		"""
		return self.dist(size=self.num)



# -*- coding: utf-8 -*-
"""
Construct pdf and cdf functions in numba for hierarchical inference.
"""
import numba
import numpy as np
from math import erf


@numba.njit
def _norm_cdf(bound,mu,sigma):  # pragma: no cover
	"""A helper function for eval_normal_logpdf_approx that calculates
	the CDF of a normal.

	Args:
		bound (float): The point at which to calculate the CDF.
		mu (float): The mean of the normal distribution
		sigma (float): The standard deviation of the normal
			distribution.

	Returns:
		(float): The CDF of the normal distribution at bound.
	"""
	return 0.5*erf((bound-mu)/(sigma*np.sqrt(2)))


@numba.njit
def eval_normal_logpdf_approx(eval_at, mu, sigma, lower=-np.inf,
	upper=np.inf):  # pragma: no cover
	"""Evaluate the log of the normal pdf, optionally truncated.

	Args:
		eval_at (np.array): The points at which to evaluate the log pdf.
		mu (float): The mean of the normal distribution
		sigma (float): The standard deviation of the normal distribution
		lower (float): If not -np.inf, the lower bound of the normal
			distribution
		upper (float): If not np.inf, the upper bound of the normal
			distribution.
	"""
	# First calculate the function without bounds
	norm = -np.log(sigma)-np.log(2*np.pi)/2
	eval_logpdf = -np.power((eval_at-mu)/sigma,2)/2+norm
	accept_norm = _norm_cdf(upper,mu,sigma) - _norm_cdf(lower,mu,sigma)

	# Now correct for the bounds if they are not -np.inf and np.inf
	# Note, reshaping must always be done regardless of bounds or numba will
	# not compile
	eval_shape = eval_at.shape
	eval_at = eval_at.reshape(-1)
	eval_logpdf=eval_logpdf.reshape(-1)
	if lower > -np.inf and upper < np.inf:
		for e_i in range(len(eval_at)):
			if eval_at[e_i] < lower:
				eval_logpdf[e_i] -= 1000
			if eval_at[e_i] > upper:
				eval_logpdf[e_i] -= 1000
	eval_logpdf=eval_logpdf.reshape(eval_shape)
	return eval_logpdf - np.log(accept_norm)


@numba.njit
def _lognorm_cdf(bound,mu,sigma):  # pragma: no cover
	"""A helper function for eval_lognormal_logpdf_approx that calculates
	the CDF of a lognormal.

	Args:
		bound (float): The point at which to calculate the CDF.
		mu (float): The mean of the log normal distribution
		sigma (float): The standard deviation of the log normal
			distribution.

	Returns:
		(float): The CDF of the normal distribution at bound.
	"""
	return 0.5*erf((np.log(bound)-mu)/(np.sqrt(2)*sigma))


@numba.njit
def eval_lognormal_logpdf_approx(eval_at, mu, sigma, lower=0,
	upper=np.inf):  # pragma: no cover
	"""Evaluate the log of the lognormal pdf, optionally truncated

	Args:
		eval_at (np.array): The points at which to evaluate the log pdf.
		mu (float): The mean of the lognormal distribution
		sigma (float): The standard deviation of the lognormal distribution
		lower (float): If not 0, the lower bound of the lognormal
			distribution
		upper (float): If not np.inf, the upper bound of the lognormal
			distribution.
	"""
	# First calculate the distribution without the bounds
	norm = -np.log(sigma) - np.log(eval_at) - np.log(2*np.pi)/2
	eval_unnormed_logpdf = -np.square(np.log(eval_at)-mu)/(2*sigma**2)
	eval_unnormed_logpdf += norm

	# Stop cdf from crashing if lower bound is below 0
	if lower<0:
		lower=0

	accept_norm = _lognorm_cdf(upper,mu,sigma) - _lognorm_cdf(lower,mu,sigma)
	eval_normed_logpdf = eval_unnormed_logpdf - np.log(accept_norm)

	# Now correct for the bounds if they are not -np.inf and np.inf
	# Note, reshaping must always be done regardless of bounds or numba will
	# not compile
	eval_shape = eval_at.shape
	eval_at = eval_at.reshape(-1)
	eval_normed_logpdf=eval_normed_logpdf.reshape(-1)
	if lower > 0 and upper < np.inf:
		for e_i in range(len(eval_at)):
			if eval_at[e_i] < lower:
				eval_normed_logpdf[e_i] -= 1000
			if eval_at[e_i] > upper:
				eval_normed_logpdf[e_i] -= 1000
	eval_normed_logpdf=eval_normed_logpdf.reshape(eval_shape)
	return eval_normed_logpdf



from paltas.Configs.xxxx_yyyy.config_val import *

config_dict['subhalo']['parameters']['sigma_sub'] = norm(loc=2.0e-3,
	scale=1.5e-4).rvs
config_dict['main_deflector']['parameters']['gamma'] = truncnorm(-198.8,
	np.inf,loc=1.988,scale=0.01).rvs
config_dict['main_deflector']['parameters']['theta_E'] = truncnorm(-75.2,
	np.inf,loc=1.128,scale=0.015).rvs
config_dict['main_deflector']['parameters']['e1'] = norm(loc=0.050,
	scale=0.01).rvs
config_dict['main_deflector']['parameters']['e2'] = norm(loc=0.031,
	scale=0.01).rvs
config_dict['main_deflector']['parameters']['center_x'] = norm(loc=0.002,
	scale=0.016).rvs
config_dict['main_deflector']['parameters']['center_y'] = norm(loc=-0.006,
	scale=0.016).rvs
config_dict['main_deflector']['parameters']['gamma1'] = norm(loc=0.018,
	scale=0.005).rvs
config_dict['main_deflector']['parameters']['gamma2'] = norm(loc=-0.022,
	scale=0.005).rvs



#!/usr/bin/env python
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

from setuptools import find_packages
import os

readme = open("README.rst").read()

PACKAGE_PATH = os.path.abspath(os.path.join(__file__, os.pardir))

required_packages = open('requirements.txt').read().splitlines()

setup(
	name='paltas',
	version='0.1.1',
	description='Strong lens substructure package.',
	long_description=readme,
	author='Sebastian Wagner-Carena',
	author_email='sebaswagner@outlook.com',
	url='https://github.com/swagnercarena/paltas',
	packages=find_packages(PACKAGE_PATH),
	package_dir={'paltas': 'paltas'},
	include_package_data=True,
	install_requires=required_packages,
	license='MIT',
	zip_safe=False
)



# -*- coding: utf-8 -*-
"""
Provide the implementation of a few main deflector profiles that consist
of two or fewer components.

This module contains the classes to render main deflectors consisting of simple
combinations of lenstronomy profiles.
"""
from .main_deflector_base import MainDeflectorBase
from lenstronomy.LensModel.profile_list_base import ProfileListBase


class PEMD(MainDeflectorBase):
	"""Class for rendering a main deflector the includes a PEMD profile.

	Args:
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
	Notes:
		Uses the lenstronomy EPL_NUMBA class, which is equivalent to PEMD but is
		pure python.

	Required Parameters

	- gamma - power law slope
	- theta_E - Einstein radius of the profile in units of arcseconds
	- e1 - x-direction ellipticity eccentricity
	- e2 - xy-direction ellipticity eccentricity
	- center_x - x-coordinate lens center in units of arcseconds
	- center_y - y-coordinate lens center in units of arcseconds
	- z_lens - main deflector redshift
	"""
	# Define the parameters we expect to find for the DG_19 model
	required_parameters = ('gamma','theta_E','e1','e2','center_x',
		'center_y','z_lens')

	def __init__(self,main_deflector_parameters,cosmology_parameters):

		# Initialize the super class
		super().__init__(main_deflector_parameters,cosmology_parameters)

	def draw_main_deflector(self):
		"""Draws the lenstronomy profile names and kwargs for the components
		of the main deflector.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each component of the main deflector, the second is the
			lenstronomy kwargs for each component, and the third is the
			redshift for each component.
		"""
		# The lists of model parameters, kwargs, and redshifts can all be
		# pulled fairly directly from the main_deflecctor_parameters
		md_model_list = ['EPL_NUMBA']
		md_kwargs_list = []
		md_z_list = [self.main_deflector_parameters['z_lens']] * len(
			md_model_list)

		# Use lenstronomy to sort the parameters
		for model in md_model_list:
			# The list of parameters linked to that lenstronomy model
			p_names = ProfileListBase._import_class(model,None,None).param_names
			model_kwargs = {}
			for param in p_names:
				model_kwargs[param] = (self.main_deflector_parameters[param])
			md_kwargs_list += [model_kwargs]

		return md_model_list, md_kwargs_list, md_z_list


class PEMDShear(MainDeflectorBase):
	"""Class for rendering a main deflector the includes a PEMD profile and
	external shear.

	Args:
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
	Notes:
		Uses the lenstronomy EPL class, which is equivalent to PEMD but is pure
		python.

	Required Parameters

	- gamma - power law slope
	- theta_E - Einstein radius of the profile in units of arcseconds
	- e1 - x-direction ellipticity eccentricity
	- e2 - xy-direction ellipticity eccentricity
	- center_x - x-coordinate lens center in units of arcseconds
	- center_y - y-coordinate lens center in units of arcseconds
	- gamma1 - x-direction shear
	- gamma2 - xy-direction shear
	- ra_0 - ra origin of shear in units of arcseconds
	- dec_0 - dec origin of shear in units of arcseconds
	- z_lens - main deflector redshift
	"""
	# Define the parameters we expect to find for the DG_19 model
	required_parameters = ('gamma','theta_E','e1','e2','center_x',
		'center_y','gamma1','gamma2','ra_0','dec_0','z_lens')

	def __init__(self,main_deflector_parameters,cosmology_parameters):

		# Initialize the super class
		super().__init__(main_deflector_parameters,cosmology_parameters)

	def draw_main_deflector(self):
		"""Draws the lenstronomy profile names and kwargs for the components
		of the main deflector.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each component of the main deflector, the second is the
			lenstronomy kwargs for each component, and the third is the
			redshift for each component.
		"""
		# The lists of model parameters, kwargs, and redshifts can all be
		# pulled fairly directly from the main_deflecctor_parameters
		md_model_list = ['EPL_NUMBA','SHEAR']
		md_kwargs_list = []
		md_z_list = [self.main_deflector_parameters['z_lens']] * len(
			md_model_list)

		# Use lenstronomy to sort the parameters
		for model in md_model_list:
			# The list of parameters linked to that lenstronomy model
			p_names = ProfileListBase._import_class(model,None,None).param_names
			model_kwargs = {}
			for param in p_names:
				model_kwargs[param] = (self.main_deflector_parameters[param])
			md_kwargs_list += [model_kwargs]

		return md_model_list, md_kwargs_list, md_z_list


class PEMDShearFourMultipole(PEMDShear):
	"""Class for rendering a main deflector the includes a PEMD profile,
	external shear, and order 2, 3, and 4 multipole.
	Args:
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
	Notes:
		Uses the lenstronomy EPL class, which is equivalent to PEMD but is pure
		python.
	Required Parameters
	- gamma - power law slope
	- theta_E - Einstein radius of the profile in units of arcseconds
	- e1 - x-direction ellipticity eccentricity
	- e2 - xy-direction ellipticity eccentricity
	- center_x - x-coordinate lens center in units of arcseconds
	- center_y - y-coordinate lens center in units of arcseconds
	- gamma1 - x-direction shear
	- gamma2 - xy-direction shear
	- ra_0 - ra origin of shear in units of arcseconds
	- dec_0 - dec origin of shear in units of arcseconds
	- z_lens - main deflector redshift
	- mult2_a - multipole strength for order 2 multpole
	- mult2_phi - multipole order 2 orientation in radians
	- mult2_center_x - x-coordinate multipole order 2 center in units of
	arcseconds
	- mult2_center_y - y-coordinate multipole order 2 center in units of
	arcseconds
	- mult3_a - multipole strength for order 2 multpole
	- mult3_phi - multipole order 2 orientation in radians
	- mult3_center_x - x-coordinate multipole order 2 center in units of
	arcseconds
	- mult3_center_y - y-coordinate multipole order 2 center in units of
	arcseconds
	- mult4_a - multipole strength for order 2 multpole
	- mult4_phi - multipole order 2 orientation in radians
	- mult4_center_x - x-coordinate multipole order 2 center in units of
	arcseconds
	- mult4_center_y - y-coordinate multipole order 2 center in units of
	arcseconds
	"""
	# Define the parameters we expect to find for the DG_19 model
	required_parameters = PEMDShear.required_parameters + ('mult2_a',
		'mult2_phi','mult2_center_x','mult2_center_y','mult3_a','mult3_phi',
		'mult3_center_x','mult3_center_y','mult4_a','mult4_phi',
		'mult4_center_x','mult4_center_y')

	def __init__(self,main_deflector_parameters,cosmology_parameters):

		# Initialize the super class
		super().__init__(main_deflector_parameters,cosmology_parameters)

	def draw_main_deflector(self):
		"""Draws the lenstronomy profile names and kwargs for the components
		of the main deflector.
		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each component of the main deflector, the second is the
			lenstronomy kwargs for each component, and the third is the
			redshift for each component.
		"""
		# Get the first set of lists directly from the inherited class
		md_model_list, md_kwargs_list, md_z_list = (
			super().draw_main_deflector())
		# Now add the three multipole models.
		for mult_i in range(2,5):
			# Only add multipole models with non zero strength.
			if self.main_deflector_parameters['mult%d_a'%(mult_i)] > 0:
				md_model_list += ['MULTIPOLE']
				# Initialize the model kwargs for this multipole model
				model_kwargs = {'m':mult_i,
					'a_m':self.main_deflector_parameters['mult%d_a'%(mult_i)],
					'phi_m':self.main_deflector_parameters[
						'mult%d_phi'%(mult_i)],
					'center_x':self.main_deflector_parameters[
						'mult%d_center_x'%(mult_i)],
					'center_y':self.main_deflector_parameters[
						'mult%d_center_y'%(mult_i)]}
				md_kwargs_list += [model_kwargs]
				md_z_list += [self.main_deflector_parameters['z_lens']]

		return md_model_list, md_kwargs_list, md_z_list



# -*- coding: utf-8 -*-
"""
Define the base class to draw subhalos for a lens

This module contains the base class that all the subhalo classes will build
from. Because the steps for rendering subhalos can vary between different
models, the required functions are very sparse.
"""
from ..Utils.cosmology_utils import get_cosmology
import copy


class SubhalosBase():
	""" Base class for rendering the subhalos of a main halo.

	Args:
		subhalo_parameters (dict): A dictionary containing the type of
			subhalo distribution and the value for each of its parameters.
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		source_parameters (dict): A dictionary containing the type of the
			source and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
	"""

	required_parameters = tuple()

	def __init__(self,subhalo_parameters,main_deflector_parameters,
		source_parameters,cosmology_parameters):

		# Save the parameters as a copy to avoid user confusion
		self.subhalo_parameters = copy.deepcopy(subhalo_parameters)
		self.main_deflector_parameters = copy.deepcopy(
			main_deflector_parameters)
		self.source_parameters = copy.deepcopy(source_parameters)

		# Turn our cosmology parameters into a colossus cosmology instance
		self.cosmo = get_cosmology(cosmology_parameters)

		# Check that all the needed parameters are present
		self.check_parameterization(self.__class__.required_parameters)

	def check_parameterization(self,required_params):
		""" Check that all the required parameters are present in the
		subhalo_parameters.

		Args:
			required_params ([str,...]): A list of strings containing the
				required parameters.
		"""
		if not all(elem in self.subhalo_parameters.keys() for
			elem in required_params):
			raise ValueError('Not all of the required parameters for the ' +
				'parameterization are present.')

	def update_parameters(self,subhalo_parameters=None,
		main_deflector_parameters=None,source_parameters=None,
		cosmology_parameters=None):
		"""Updated the class parameters

		Args:
			los_parameters (dict): A dictionary containing the type of
				los distribution and the value for each of its parameters.
			main_deflector_parameters (dict): A dictionary containing the type
				of main deflector and the value for each of its parameters.
			source_parameters (dict): A dictionary containing the type of the
				source and the value for each of its parameters.
			cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
				Either a name of colossus cosmology, a dict with 'cosmology name':
				name of colossus cosmology, an instance of colussus cosmology, or
				a dict with H0 and Om0 ( other parameters will be set to
				defaults).

		Notes:
			Use this function to update parameter values instead of
			starting a new class.
		"""
		if subhalo_parameters is not None:
			self.subhalo_parameters.update(subhalo_parameters)
		if main_deflector_parameters is not None:
			self.main_deflector_parameters.update(main_deflector_parameters)
		if source_parameters is not None:
			self.source_parameters.update(source_parameters)
		if cosmology_parameters is not None:
			self.cosmo = get_cosmology(cosmology_parameters)

	def draw_subhalos(self):
		"""Draws masses, concentrations,and positions for the subhalos of a
		main lens halo.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each subhalo returned, the second is the lenstronomy kwargs for
			that subhalo, and the third is the redshift for each subhalo.
		"""
		raise NotImplementedError
